import arvados
import arvados.util
import arvados.collection
import arvados.errors
import os
import connexion
import json
import subprocess
import tempfile
import functools
import threading
import logging
import shutil

from wes_service.util import visit, WESBackend


class MissingAuthorization(Exception):
    pass


def get_api(authtoken=None):
    if authtoken is None:
        if not connexion.request.headers.get("Authorization"):
            raise MissingAuthorization()
        authtoken = connexion.request.headers["Authorization"]
        if not authtoken.startswith("Bearer ") or authtoken.startswith("OAuth2 "):
            raise ValueError("Authorization token must start with 'Bearer '")
        authtoken = authtoken[7:]
    return arvados.api_from_config(
        version="v1",
        apiconfig={
            "ARVADOS_API_HOST": os.environ["ARVADOS_API_HOST"],
            "ARVADOS_API_TOKEN": authtoken,
            "ARVADOS_API_HOST_INSECURE": os.environ.get(
                "ARVADOS_API_HOST_INSECURE", "false"
            ),  # NOQA
        },
    )


statemap = {
    "Queued": "QUEUED",
    "Locked": "INITIALIZING",
    "Running": "RUNNING",
    "Complete": "COMPLETE",
    "Cancelled": "CANCELED",
}


def catch_exceptions(orig_func):
    """Catch uncaught exceptions and turn them into http errors"""

    @functools.wraps(orig_func)
    def catch_exceptions_wrapper(self, *args, **kwargs):
        try:
            return orig_func(self, *args, **kwargs)
        except arvados.errors.ApiError as e:
            logging.exception("Failure")
            return (
                {"msg": e._get_reason(), "status_code": e.resp.status},
                int(e.resp.status),
            )
        except subprocess.CalledProcessError as e:
            return {"msg": str(e), "status_code": 500}, 500
        except MissingAuthorization:
            return (
                {
                    "msg": "'Authorization' header is missing or empty, expecting Arvados API token",
                    "status_code": 401,
                },
                401,
            )
        except ValueError as e:
            return {"msg": str(e), "status_code": 400}, 400
        except Exception as e:
            return {"msg": str(e), "status_code": 500}, 500

    return catch_exceptions_wrapper


class ArvadosBackend(WESBackend):
    def GetServiceInfo(self):
        stdout, stderr = subprocess.Popen(
            ["arvados-cwl-runner", "--version"], stderr=subprocess.PIPE
        ).communicate()
        return {
            "workflow_type_versions": {
                "CWL": {"workflow_type_version": ["v1.0", "v1.1", "v1.2"]}
            },
            "supported_wes_versions": ["0.3.0", "1.0.0"],
            "supported_filesystem_protocols": ["http", "https", "keep"],
            "workflow_engine_versions": {"arvados-cwl-runner": str(stderr)},
            "default_workflow_engine_parameters": [],
            "system_state_counts": {},
            "auth_instructions_url": "http://doc.arvados.org/user/reference/api-tokens.html",
            "tags": {"ARVADOS_API_HOST": os.environ["ARVADOS_API_HOST"]},
        }

    @catch_exceptions
    def ListRuns(self, page_size=None, page_token=None, state_search=None):
        api = get_api()

        paging = []
        if page_token:
            paging = [["uuid", ">", page_token]]

        requests = (
            api.container_requests()
            .list(
                filters=[
                    ["requesting_container_uuid", "=", None],
                    ["container_uuid", "!=", None],
                ]
                + paging,
                select=["uuid", "command", "container_uuid"],
                order=["uuid"],
                limit=page_size,
            )
            .execute()["items"]
        )
        containers = (
            api.containers()
            .list(
                filters=[["uuid", "in", [w["container_uuid"] for w in requests]]],
                select=["uuid", "state"],
            )
            .execute()["items"]
        )

        uuidmap = {c["uuid"]: statemap[c["state"]] for c in containers}

        workflow_list = [
            {"run_id": cr["uuid"], "state": uuidmap.get(cr["container_uuid"])}
            for cr in requests
            if cr["command"] and cr["command"][0] == "arvados-cwl-runner"
        ]
        return {
            "workflows": workflow_list,
            "next_page_token": workflow_list[-1]["run_id"] if workflow_list else "",
        }

    def log_for_run(self, run_id, message, authtoken=None):
        get_api(authtoken).logs().create(
            body={
                "log": {
                    "object_uuid": run_id,
                    "event_type": "stderr",
                    "properties": {"text": message + "\n"},
                }
            }
        ).execute()

    def invoke_cwl_runner(
        self, cr_uuid, workflow_url, workflow_params, env, project_uuid, tempdir
    ):
        api = arvados.api_from_config(
            version="v1",
            apiconfig={
                "ARVADOS_API_HOST": env["ARVADOS_API_HOST"],
                "ARVADOS_API_TOKEN": env["ARVADOS_API_TOKEN"],
                "ARVADOS_API_HOST_INSECURE": env["ARVADOS_API_HOST_INSECURE"],  # NOQA
            },
        )

        try:
            with tempfile.NamedTemporaryFile(
                "wt", dir=tempdir, suffix=".json"
            ) as inputtemp:
                json.dump(workflow_params, inputtemp)
                inputtemp.flush()

                msg = ""
                for dirpath, dirs, files in os.walk(tempdir):
                    for f in files:
                        msg += "  " + dirpath + "/" + f + "\n"

                self.log_for_run(
                    cr_uuid,
                    f"Contents of {tempdir}:\n{msg}",
                    env["ARVADOS_API_TOKEN"],
                )

                # TODO: run submission process in a container to prevent
                # a-c-r submission processes from seeing each other.

                cmd = [
                    "arvados-cwl-runner",
                    "--submit-request-uuid=" + cr_uuid,
                    "--submit",
                    "--no-wait",
                    "--api=containers",
                    "--debug",
                ]

                if project_uuid:
                    cmd.append("--project-uuid=" + project_uuid)

                cmd.append(workflow_url)
                cmd.append(inputtemp.name)

                self.log_for_run(
                    cr_uuid, "Executing %s" % cmd, env["ARVADOS_API_TOKEN"]
                )

                proc = subprocess.Popen(
                    cmd,
                    env=env,
                    cwd=tempdir,
                    stdout=subprocess.PIPE,
                    stderr=subprocess.PIPE,
                )
                (stdoutdata, stderrdata) = proc.communicate()
                if proc.returncode != 0:
                    api.container_requests().update(
                        uuid=cr_uuid, body={"priority": 0}
                    ).execute()

                self.log_for_run(
                    cr_uuid, stderrdata.decode("utf-8"), env["ARVADOS_API_TOKEN"]
                )

            if tempdir:
                shutil.rmtree(tempdir)

        except subprocess.CalledProcessError as e:
            api.container_requests().update(
                uuid=cr_uuid,
                body={
                    "priority": 0,
                    "name": "Cancelled container request",
                    "properties": {"arvados-cwl-runner-log": str(e)},
                },
            ).execute()

    @catch_exceptions
    def RunWorkflow(self, **args):
        if not connexion.request.headers.get("Authorization"):
            raise MissingAuthorization()

        authtoken = connexion.request.headers["Authorization"]
        if authtoken.startswith("Bearer ") or authtoken.startswith("OAuth2 "):
            authtoken = authtoken[7:]

        env = {
            "PATH": os.environ["PATH"],
            "ARVADOS_API_HOST": os.environ["ARVADOS_API_HOST"],
            "ARVADOS_API_TOKEN": authtoken,
            "ARVADOS_API_HOST_INSECURE": os.environ.get(
                "ARVADOS_API_HOST_INSECURE", "false"
            ),  # NOQA
        }

        api = get_api()

        cr = (
            api.container_requests()
            .create(
                body={
                    "container_request": {
                        "command": [""],
                        "container_image": "n/a",
                        "state": "Uncommitted",
                        "output_path": "n/a",
                        "priority": 500,
                    }
                }
            )
            .execute()
        )

        try:
            tempdir, body = self.collect_attachments(cr["uuid"])

            workflow_engine_parameters = body.get("workflow_engine_parameters", {})
            project_uuid = None
            if workflow_engine_parameters:
                project_uuid = workflow_engine_parameters.get("project_uuid")

            threading.Thread(
                target=self.invoke_cwl_runner,
                args=(
                    cr["uuid"],
                    body["workflow_url"],
                    body["workflow_params"],
                    env,
                    project_uuid,
                    tempdir,
                ),
            ).start()
        except ValueError as e:
            self.log_for_run(cr["uuid"], "Bad request: " + str(e))
            cr = (
                api.container_requests()
                .update(
                    uuid=cr["uuid"],
                    body={
                        "container_request": {
                            "name": "Cancelled container request",
                            "priority": 0,
                        }
                    },
                )
                .execute()
            )
            return {"msg": str(e), "status_code": 400}, 400
        except Exception as e:
            logging.exception("Error")
            self.log_for_run(
                cr["uuid"],
                "An exception ocurred while handling your request: " + str(e),
            )
            cr = (
                api.container_requests()
                .update(
                    uuid=cr["uuid"],
                    body={
                        "container_request": {
                            "name": "Cancelled container request",
                            "priority": 0,
                        }
                    },
                )
                .execute()
            )
            return {"msg": str(e), "status_code": 500}, 500
        else:
            return {"run_id": cr["uuid"]}

    @catch_exceptions
    def GetRunLog(self, run_id):
        api = get_api()

        request = api.container_requests().get(uuid=run_id).execute()
        if request["container_uuid"]:
            container = (
                api.containers().get(uuid=request["container_uuid"]).execute()
            )  # NOQA
            task_reqs = arvados.util.list_all(
                api.container_requests().list,
                filters=[["requesting_container_uuid", "=", container["uuid"]]],
            )
            tasks = arvados.util.list_all(
                api.containers().list,
                filters=[["uuid", "in", [tr["container_uuid"] for tr in task_reqs]]],
            )
            containers_map = {c["uuid"]: c for c in tasks}
            containers_map[container["uuid"]] = container
        else:
            container = {
                "state": "Queued" if request["priority"] > 0 else "Cancelled",
                "exit_code": None,
                "log": None,
            }
            tasks = []
            containers_map = {}
            task_reqs = []

        outputobj = {}
        if request["output_uuid"]:
            c = arvados.collection.CollectionReader(
                request["output_uuid"], api_client=api
            )
            with c.open("cwl.output.json") as f:
                try:
                    outputobj = json.load(f)
                except ValueError:
                    pass

                def keepref(d):
                    if isinstance(d, dict) and "location" in d:
                        d["location"] = "{}c={}/_/{}".format(
                            api._resourceDesc["keepWebServiceUrl"],
                            c.portable_data_hash(),
                            d["location"],
                        )  # NOQA

                visit(outputobj, keepref)

        def log_object(cr):
            if cr["container_uuid"]:
                containerlog = containers_map[cr["container_uuid"]]
            else:
                containerlog = {
                    "started_at": "",
                    "finished_at": "",
                    "exit_code": None,
                    "log": "",
                }
            r = {
                "name": cr["name"] or "",
                "cmd": cr["command"],
                "start_time": containerlog["started_at"] or "",
                "end_time": containerlog["finished_at"] or "",
                "stdout": "",
                "stderr": "",
                "exit_code": containerlog["exit_code"] or 0,
            }
            if containerlog["log"]:
                r["stdout_keep"] = "{}c={}/_/{}".format(
                    api._resourceDesc["keepWebServiceUrl"],
                    containerlog["log"],
                    "stdout.txt",
                )  # NOQA
                r["stderr_keep"] = "{}c={}/_/{}".format(
                    api._resourceDesc["keepWebServiceUrl"],
                    containerlog["log"],
                    "stderr.txt",
                )  # NOQA
            r["stdout"] = "%s/x-dynamic-logs/stdout" % (connexion.request.url)
            r["stderr"] = "%s/x-dynamic-logs/stderr" % (connexion.request.url)

            return r

        r = {
            "run_id": request["uuid"],
            "request": {
                "workflow_url": "",
                "workflow_params": request["mounts"]
                .get("/var/lib/cwl/cwl.input.json", {})
                .get("content", {}),
            },
            "state": statemap[container["state"]],
            "run_log": log_object(request),
            "task_logs": [log_object(t) for t in task_reqs],
            "outputs": outputobj,
        }

        return r

    @catch_exceptions
    def CancelRun(self, run_id):  # NOQA
        api = get_api()
        request = (
            api.container_requests().update(uuid=run_id, body={"priority": 0}).execute()
        )  # NOQA
        return {"run_id": request["uuid"]}

    @catch_exceptions
    def GetRunStatus(self, run_id):
        api = get_api()
        request = api.container_requests().get(uuid=run_id).execute()
        if request["container_uuid"]:
            container = (
                api.containers().get(uuid=request["container_uuid"]).execute()
            )  # NOQA
        elif request["priority"] == 0:
            container = {"state": "Cancelled"}
        else:
            container = {"state": "Queued"}
        return {"run_id": request["uuid"], "state": statemap[container["state"]]}


def dynamic_logs(run_id, logstream):
    api = get_api()
    cr = api.container_requests().get(uuid=run_id).execute()
    l1 = [
        t["properties"]["text"]
        for t in api.logs()
        .list(
            filters=[["object_uuid", "=", run_id], ["event_type", "=", logstream]],
            order="created_at desc",
            limit=100,
        )
        .execute()["items"]
    ]
    if cr["container_uuid"]:
        l2 = [
            t["properties"]["text"]
            for t in api.logs()
            .list(
                filters=[
                    ["object_uuid", "=", cr["container_uuid"]],
                    ["event_type", "=", logstream],
                ],
                order="created_at desc",
                limit=100,
            )
            .execute()["items"]
        ]
    else:
        l2 = []
    return "".join(reversed(l1)) + "".join(reversed(l2))


def create_backend(app, opts):
    ab = ArvadosBackend(opts)
    app.app.route("/ga4gh/wes/v1/runs/<run_id>/x-dynamic-logs/<logstream>")(
        dynamic_logs
    )
    return ab



#!/usr/bin/env python
import argparse
import pkg_resources  # part of setuptools
import sys
import ruamel.yaml
import os
import logging
import connexion
import connexion.utils as utils
from connexion.resolver import Resolver

logging.basicConfig(level=logging.INFO)


def setup(args=None):
    if args is None:
        args = get_parser().parse_args([])  # grab the defaults

    configfile = "config.yml"
    if os.path.isfile(configfile):
        logging.info("Loading %s", configfile)
        with open(configfile) as f:
            config = ruamel.yaml.safe_load(f)
        for c in config:
            setattr(args, c, config[c])

    logging.info("Using config:")
    for n in args.__dict__:
        logging.info("  %s: %s", n, getattr(args, n))

    app = connexion.App(__name__)
    backend = utils.get_function_from_name(args.backend + ".create_backend")(
        app, args.opt
    )

    def rs(x):
        return getattr(backend, x.split(".")[-1])

    app.add_api(
        "openapi/workflow_execution_service.swagger.yaml", resolver=Resolver(rs)
    )

    return app


def get_parser() -> argparse.Namespace:
    parser = argparse.ArgumentParser(description="Workflow Execution Service")
    parser.add_argument(
        "--backend",
        type=str,
        default="wes_service.cwl_runner",
        help="Either: '--backend=wes_service.arvados_wes' or '--backend=wes_service.cwl_runner'",
    )
    parser.add_argument("--port", type=int, default=8080)
    parser.add_argument(
        "--opt",
        type=str,
        action="append",
        help="Example: '--opt runner=cwltoil --opt extra=--logLevel=CRITICAL' "
        "or '--opt extra=--workDir=/'.  Accepts multiple values.",
    )
    parser.add_argument("--debug", action="store_true", default=False)
    parser.add_argument("--version", action="store_true", default=False)
    return parser


def main(argv=sys.argv[1:]):
    args = get_parser().parse_args(argv)

    if args.version:
        pkg = pkg_resources.require("wes_service")
        print(f"{sys.argv[0]} {pkg[0].version}")
        exit(0)

    app = setup(args)

    app.run(port=args.port, debug=args.debug)


if __name__ == "__main__":
    main(sys.argv[1:])



import json
import os
import subprocess
import time
import logging
import uuid
import shutil

from multiprocessing import Process
from wes_service.util import WESBackend

logging.basicConfig(level=logging.INFO)


class ToilWorkflow:
    def __init__(self, run_id):
        """
        Represents a toil workflow.

        :param str run_id: A uuid string.  Used to name the folder that contains
            all of the files containing this particular workflow instance's information.
        """
        super().__init__()
        self.run_id = run_id

        self.workdir = os.path.join(os.getcwd(), "workflows", self.run_id)
        self.outdir = os.path.join(self.workdir, "outdir")
        if not os.path.exists(self.outdir):
            os.makedirs(self.outdir)

        self.outfile = os.path.join(self.workdir, "stdout")
        self.errfile = os.path.join(self.workdir, "stderr")
        self.starttime = os.path.join(self.workdir, "starttime")
        self.endtime = os.path.join(self.workdir, "endtime")
        self.pidfile = os.path.join(self.workdir, "pid")
        self.statcompletefile = os.path.join(self.workdir, "status_completed")
        self.staterrorfile = os.path.join(self.workdir, "status_error")
        self.cmdfile = os.path.join(self.workdir, "cmd")
        self.jobstorefile = os.path.join(self.workdir, "jobstore")
        self.request_json = os.path.join(self.workdir, "request.json")
        self.input_json = os.path.join(self.workdir, "wes_input.json")
        self.jobstore_default = "file:" + os.path.join(self.workdir, "toiljobstore")
        self.jobstore = None

    def sort_toil_options(self, extra):
        # determine jobstore and set a new default if the user did not set one
        cloud = False
        for e in extra:
            if e.startswith("--jobStore="):
                self.jobstore = e[11:]
                if self.jobstore.startswith(("aws", "google", "azure")):
                    cloud = True
            if e.startswith(("--outdir=", "-o=")):
                extra.remove(e)
        if not cloud:
            extra.append("--outdir=" + self.outdir)
        if not self.jobstore:
            extra.append("--jobStore=" + self.jobstore_default)
            self.jobstore = self.jobstore_default

        # store the jobstore location
        with open(self.jobstorefile, "w") as f:
            f.write(self.jobstore)

        return extra

    def write_workflow(self, request, opts, cwd, wftype="cwl"):
        """Writes a cwl, wdl, or python file as appropriate from the request dictionary."""

        workflow_url = request.get("workflow_url")

        # link the cwl and json into the cwd
        if workflow_url.startswith("file://"):
            try:
                os.link(workflow_url[7:], os.path.join(cwd, "wes_workflow." + wftype))
            except OSError:
                os.symlink(
                    workflow_url[7:], os.path.join(cwd, "wes_workflow." + wftype)
                )
            workflow_url = os.path.join(cwd, "wes_workflow." + wftype)
        try:
            os.link(self.input_json, os.path.join(cwd, "wes_input.json"))
        except OSError:
            os.symlink(self.input_json, os.path.join(cwd, "wes_input.json"))
        self.input_json = os.path.join(cwd, "wes_input.json")

        extra_options = self.sort_toil_options(opts.getoptlist("extra"))
        if wftype == "cwl":
            command_args = (
                ["toil-cwl-runner"] + extra_options + [workflow_url, self.input_json]
            )
        elif wftype == "wdl":
            command_args = (
                ["toil-wdl-runner"] + extra_options + [workflow_url, self.input_json]
            )
        elif wftype == "py":
            command_args = ["python"] + extra_options + [workflow_url]
        else:
            raise RuntimeError(
                'workflow_type is not "cwl", "wdl", or "py": ' + str(wftype)
            )

        return command_args

    def write_json(self, request_dict):
        input_json = os.path.join(self.workdir, "input.json")
        with open(input_json, "w") as f:
            json.dump(request_dict["workflow_params"], f)
        return input_json

    def call_cmd(self, cmd, cwd):
        """
        Calls a command with Popen.
        Writes stdout, stderr, and the command to separate files.

        :param cmd: A string or array of strings.
        :param tempdir:
        :return: The pid of the command.
        """
        with open(self.cmdfile, "w") as f:
            f.write(str(cmd))
        stdout = open(self.outfile, "w")
        stderr = open(self.errfile, "w")
        logging.info(
            "Calling: %s, with outfile: %s and errfile: %s",
            (" ".join(cmd)),
            self.outfile,
            self.errfile,
        )
        process = subprocess.Popen(
            cmd, stdout=stdout, stderr=stderr, close_fds=True, cwd=cwd
        )
        stdout.close()
        stderr.close()

        return process.pid

    def cancel(self):
        pass

    def fetch(self, filename):
        if os.path.exists(filename):
            with open(filename) as f:
                return f.read()
        return ""

    def getlog(self):
        state, exit_code = self.getstate()

        with open(self.request_json) as f:
            request = json.load(f)

        with open(self.jobstorefile) as f:
            self.jobstore = f.read()

        stderr = self.fetch(self.errfile)
        starttime = self.fetch(self.starttime)
        endtime = self.fetch(self.endtime)
        cmd = [self.fetch(self.cmdfile)]

        outputobj = {}
        if state == "COMPLETE":
            # only tested locally
            if self.jobstore.startswith("file:"):
                for f in os.listdir(self.outdir):
                    if f.startswith("out_tmpdir"):
                        shutil.rmtree(os.path.join(self.outdir, f))
                for f in os.listdir(self.outdir):
                    outputobj[f] = {
                        "location": os.path.join(self.outdir, f),
                        "size": os.stat(os.path.join(self.outdir, f)).st_size,
                        "class": "File",
                    }

        return {
            "run_id": self.run_id,
            "request": request,
            "state": state,
            "run_log": {
                "cmd": cmd,
                "start_time": starttime,
                "end_time": endtime,
                "stdout": "",
                "stderr": stderr,
                "exit_code": exit_code,
            },
            "task_logs": [],
            "outputs": outputobj,
        }

    def run(self, request, tempdir, opts):
        """
        Constructs a command to run a cwl/json from requests and opts,
        runs it, and deposits the outputs in outdir.

        Runner:
        opts.getopt("runner", default="cwl-runner")

        CWL (url):
        request["workflow_url"] == a url to a cwl file
        or
        request["workflow_attachment"] == input cwl text (written to a file and a url constructed for that file)

        JSON File:
        request["workflow_params"] == input json text (to be written to a file)

        :param dict request: A dictionary containing the cwl/json information.
        :param str tempdir: Folder where input files have been staged and the cwd to run at.
        :param wes_service.util.WESBackend opts: contains the user's arguments;
                                                 specifically the runner and runner options
        :return: {"run_id": self.run_id, "state": state}
        """
        wftype = request["workflow_type"].lower().strip()
        version = request["workflow_type_version"]

        if wftype == "cwl" and version not in ("v1.0", "v1.1", "v1.2"):
            raise RuntimeError(
                'workflow_type "cwl" requires '
                '"workflow_type_version" to be "v1.[012]": ' + str(version)
            )
        if version != "2.7" and wftype == "py":
            raise RuntimeError(
                'workflow_type "py" requires '
                '"workflow_type_version" to be "2.7": ' + str(version)
            )

        logging.info("Beginning Toil Workflow ID: " + str(self.run_id))

        with open(self.starttime, "w") as f:
            f.write(str(time.time()))
        with open(self.request_json, "w") as f:
            json.dump(request, f)
        with open(self.input_json, "w") as inputtemp:
            json.dump(request["workflow_params"], inputtemp)

        command_args = self.write_workflow(request, opts, tempdir, wftype=wftype)
        pid = self.call_cmd(command_args, tempdir)

        with open(self.endtime, "w") as f:
            f.write(str(time.time()))
        with open(self.pidfile, "w") as f:
            f.write(str(pid))

        return self.getstatus()

    def getstate(self):
        """
        Returns QUEUED,          -1
                INITIALIZING,    -1
                RUNNING,         -1
                COMPLETE,         0
                or
                EXECUTOR_ERROR, 255
        """
        # the jobstore never existed
        if not os.path.exists(self.jobstorefile):
            logging.info("Workflow " + self.run_id + ": QUEUED")
            return "QUEUED", -1

        # completed earlier
        if os.path.exists(self.statcompletefile):
            logging.info("Workflow " + self.run_id + ": COMPLETE")
            return "COMPLETE", 0

        # errored earlier
        if os.path.exists(self.staterrorfile):
            logging.info("Workflow " + self.run_id + ": EXECUTOR_ERROR")
            return "EXECUTOR_ERROR", 255

        # the workflow is staged but has not run yet
        if not os.path.exists(self.errfile):
            logging.info("Workflow " + self.run_id + ": INITIALIZING")
            return "INITIALIZING", -1

        completed = False
        with open(self.errfile) as f:
            for line in f:
                if "Traceback (most recent call last)" in line:
                    logging.info("Workflow " + self.run_id + ": EXECUTOR_ERROR")
                    open(self.staterrorfile, "a").close()
                    return "EXECUTOR_ERROR", 255
        if (
            subprocess.run(
                ["toil", "status", "--failIfNotComplete", self.jobstorefile]
            ).returncode
            == 0
        ):
            completed = True
        if completed:
            logging.info("Workflow " + self.run_id + ": COMPLETE")
            open(self.statcompletefile, "a").close()
            return "COMPLETE", 0

        logging.info("Workflow " + self.run_id + ": RUNNING")
        return "RUNNING", -1

    def getstatus(self):
        state, exit_code = self.getstate()

        return {"run_id": self.run_id, "state": state}


class ToilBackend(WESBackend):
    processes = {}

    def GetServiceInfo(self):
        return {
            "workflow_type_versions": {
                "CWL": {"workflow_type_version": ["v1.0", "v1.1", "v1.2"]},
                "WDL": {"workflow_type_version": ["draft-2"]},
                "PY": {"workflow_type_version": ["2.7"]},
            },
            "supported_wes_versions": ["0.3.0", "1.0.0"],
            "supported_filesystem_protocols": ["file", "http", "https"],
            "workflow_engine_versions": ["3.16.0"],
            "system_state_counts": {},
            "key_values": {},
        }

    def ListRuns(self, page_size=None, page_token=None, state_search=None):
        # FIXME #15 results don't page
        if not os.path.exists(os.path.join(os.getcwd(), "workflows")):
            return {"workflows": [], "next_page_token": ""}
        wf = []
        for entry in os.listdir(os.path.join(os.getcwd(), "workflows")):
            if os.path.isdir(os.path.join(os.getcwd(), "workflows", entry)):
                wf.append(ToilWorkflow(entry))

        workflows = [{"run_id": w.run_id, "state": w.getstate()[0]} for w in wf]  # NOQA
        return {"workflows": workflows, "next_page_token": ""}

    def RunWorkflow(self):
        tempdir, body = self.collect_attachments()

        run_id = uuid.uuid4().hex
        job = ToilWorkflow(run_id)
        p = Process(target=job.run, args=(body, tempdir, self))
        p.start()
        self.processes[run_id] = p
        return {"run_id": run_id}

    def GetRunLog(self, run_id):
        job = ToilWorkflow(run_id)
        return job.getlog()

    def CancelRun(self, run_id):
        # should this block with `p.is_alive()`?
        if run_id in self.processes:
            self.processes[run_id].terminate()
        return {"run_id": run_id}

    def GetRunStatus(self, run_id):
        job = ToilWorkflow(run_id)
        return job.getstatus()


def create_backend(app, opts):
    return ToilBackend(opts)



import json
import os
import subprocess
import uuid

from wes_service.util import WESBackend


class Workflow:
    def __init__(self, run_id):
        super().__init__()
        self.run_id = run_id
        self.workdir = os.path.join(os.getcwd(), "workflows", self.run_id)
        self.outdir = os.path.join(self.workdir, "outdir")
        if not os.path.exists(self.outdir):
            os.makedirs(self.outdir)

    def run(self, request, tempdir, opts):
        """
        Constructs a command to run a cwl/json from requests and opts,
        runs it, and deposits the outputs in outdir.

        Runner:
        opts.getopt("runner", default="cwl-runner")

        CWL (url):
        request["workflow_url"] == a url to a cwl file
        or
        request["workflow_attachment"] == input cwl text (written to a file and a url constructed for that file)

        JSON File:
        request["workflow_params"] == input json text (to be written to a file)

        :param dict request: A dictionary containing the cwl/json information.
        :param wes_service.util.WESBackend opts: contains the user's arguments;
                                                 specifically the runner and runner options
        :return: {"run_id": self.run_id, "state": state}
        """
        with open(os.path.join(self.workdir, "request.json"), "w") as f:
            json.dump(request, f)

        with open(os.path.join(self.workdir, "cwl.input.json"), "w") as inputtemp:
            json.dump(request["workflow_params"], inputtemp)

        workflow_url = request.get(
            "workflow_url"
        )  # Will always be local path to descriptor cwl, or url.

        output = open(os.path.join(self.workdir, "cwl.output.json"), "w")
        stderr = open(os.path.join(self.workdir, "stderr"), "w")

        runner = opts.getopt("runner", default="cwl-runner")
        extra = opts.getoptlist("extra")

        # replace any locally specified outdir with the default
        for e in extra:
            if e.startswith("--outdir="):
                extra.remove(e)
        extra.append("--outdir=" + self.outdir)

        # link the cwl and json into the tempdir/cwd
        if workflow_url.startswith("file://"):
            os.symlink(workflow_url[7:], os.path.join(tempdir, "wes_workflow.cwl"))
            workflow_url = os.path.join(tempdir, "wes_workflow.cwl")
        os.symlink(inputtemp.name, os.path.join(tempdir, "cwl.input.json"))
        jsonpath = os.path.join(tempdir, "cwl.input.json")

        # build args and run
        command_args = [runner] + extra + [workflow_url, jsonpath]
        proc = subprocess.Popen(
            command_args, stdout=output, stderr=stderr, close_fds=True, cwd=tempdir
        )
        output.close()
        stderr.close()
        with open(os.path.join(self.workdir, "pid"), "w") as pid:
            pid.write(str(proc.pid))

        return self.getstatus()

    def getstate(self):
        """
        Returns RUNNING, -1
                COMPLETE, 0
                or
                EXECUTOR_ERROR, 255
        """
        state = "RUNNING"
        exit_code = -1

        exitcode_file = os.path.join(self.workdir, "exit_code")
        pid_file = os.path.join(self.workdir, "pid")

        if os.path.exists(exitcode_file):
            with open(exitcode_file) as f:
                exit_code = int(f.read())
        elif os.path.exists(pid_file):
            with open(pid_file) as pid:
                pid = int(pid.read())
            try:
                (_pid, exit_status) = os.waitpid(pid, os.WNOHANG)
                if _pid != 0:
                    exit_code = exit_status >> 8
                    with open(exitcode_file, "w") as f:
                        f.write(str(exit_code))
                    os.unlink(pid_file)
            except OSError:
                os.unlink(pid_file)
                exit_code = 255

        if exit_code == 0:
            state = "COMPLETE"
        elif exit_code != -1:
            state = "EXECUTOR_ERROR"

        return state, exit_code

    def getstatus(self):
        state, exit_code = self.getstate()

        return {"run_id": self.run_id, "state": state}

    def getlog(self):
        state, exit_code = self.getstate()

        with open(os.path.join(self.workdir, "request.json")) as f:
            request = json.load(f)

        with open(os.path.join(self.workdir, "stderr")) as f:
            stderr = f.read()

        outputobj = {}
        if state == "COMPLETE":
            output_path = os.path.join(self.workdir, "cwl.output.json")
            with open(output_path) as outputtemp:
                outputobj = json.load(outputtemp)

        return {
            "run_id": self.run_id,
            "request": request,
            "state": state,
            "run_log": {
                "cmd": [""],
                "start_time": "",
                "end_time": "",
                "stdout": "",
                "stderr": stderr,
                "exit_code": exit_code,
            },
            "task_logs": [],
            "outputs": outputobj,
        }

    def cancel(self):
        pass


class CWLRunnerBackend(WESBackend):
    def GetServiceInfo(self):
        runner = self.getopt("runner", default="cwl-runner")
        stdout, stderr = subprocess.Popen(
            [runner, "--version"], stderr=subprocess.PIPE
        ).communicate()
        r = {
            "workflow_type_versions": {"CWL": {"workflow_type_version": ["v1.0", "v1.1", "v1.2"]}},
            "supported_wes_versions": ["0.3.0", "1.0.0"],
            "supported_filesystem_protocols": ["file", "http", "https"],
            "workflow_engine_versions": {"cwl-runner": str(stderr)},
            "system_state_counts": {},
            "tags": {},
        }
        return r

    def ListRuns(self, page_size=None, page_token=None, state_search=None):
        # FIXME #15 results don't page
        if not os.path.exists(os.path.join(os.getcwd(), "workflows")):
            return {"workflows": [], "next_page_token": ""}
        wf = []
        for entry in os.listdir(os.path.join(os.getcwd(), "workflows")):
            if os.path.isdir(os.path.join(os.getcwd(), "workflows", entry)):
                wf.append(Workflow(entry))

        workflows = [{"run_id": w.run_id, "state": w.getstate()[0]} for w in wf]  # NOQA
        return {"workflows": workflows, "next_page_token": ""}

    def RunWorkflow(self, **args):
        tempdir, body = self.collect_attachments()

        run_id = uuid.uuid4().hex
        job = Workflow(run_id)

        job.run(body, tempdir, self)
        return {"run_id": run_id}

    def GetRunLog(self, run_id):
        job = Workflow(run_id)
        return job.getlog()

    def CancelRun(self, run_id):
        job = Workflow(run_id)
        job.cancel()
        return {"run_id": run_id}

    def GetRunStatus(self, run_id):
        job = Workflow(run_id)
        return job.getstatus()


def create_backend(app, opts):
    return CWLRunnerBackend(opts)



#!/usr/bin/env python
import pkg_resources  # part of setuptools
import json
import time
import sys
import os
import argparse
import logging
import requests
from requests.exceptions import InvalidSchema, MissingSchema
from wes_client.util import modify_jsonyaml_paths, WESClient


def main(argv=sys.argv[1:]):
    parser = argparse.ArgumentParser(description="Workflow Execution Service")
    parser.add_argument(
        "--host",
        type=str,
        default=os.environ.get("WES_API_HOST"),
        help="Example: '--host=localhost:8080'.  Defaults to WES_API_HOST.",
    )
    parser.add_argument(
        "--auth",
        type=str,
        default=os.environ.get("WES_API_AUTH"),
        help="Format is 'Header: value' or just 'value'.  If header name is not provided, value goes in the 'Authorization'.  Defaults to WES_API_AUTH.",
    )
    parser.add_argument(
        "--proto",
        type=str,
        default=os.environ.get("WES_API_PROTO", "https"),
        help="Options: [http, https].  Defaults to WES_API_PROTO (https).",
    )
    parser.add_argument("--quiet", action="store_true", default=False)
    parser.add_argument("--outdir", type=str)
    parser.add_argument(
        "--attachments",
        type=str,
        default=None,
        help="A comma separated list of attachments to include.  Example: "
        '--attachments="testdata/dockstore-tool-md5sum.cwl,testdata/md5sum.input"',
    )
    parser.add_argument("--page", type=str, default=None)
    parser.add_argument("--page-size", type=int, default=None)

    exgroup = parser.add_mutually_exclusive_group()
    exgroup.add_argument("--run", action="store_true", default=False)
    exgroup.add_argument(
        "--get",
        type=str,
        default=None,
        help="Specify a <workflow-id>.  Example: '--get=<workflow-id>'",
    )
    exgroup.add_argument(
        "--log",
        type=str,
        default=None,
        help="Specify a <workflow-id>.  Example: '--log=<workflow-id>'",
    )
    exgroup.add_argument("--list", action="store_true", default=False)
    exgroup.add_argument("--info", action="store_true", default=False)
    exgroup.add_argument("--version", action="store_true", default=False)

    exgroup = parser.add_mutually_exclusive_group()
    exgroup.add_argument("--wait", action="store_true", default=True, dest="wait")
    exgroup.add_argument("--no-wait", action="store_false", default=True, dest="wait")

    parser.add_argument("workflow_url", type=str, nargs="?", default=None)
    parser.add_argument("job_order", type=str, nargs="?", default=None)
    args = parser.parse_args(argv)

    if args.version:
        pkg = pkg_resources.require("wes_service")
        print(f"{sys.argv[0]} {pkg[0].version}")
        exit(0)

    auth = {}
    if args.auth:
        if ": " in args.auth:
            sp = args.auth.split(": ")
            auth[sp[0]] = sp[1]
        else:
            auth["Authorization"] = args.auth

    client = WESClient({"auth": auth, "proto": args.proto, "host": args.host})

    if args.list:
        response = (
            client.list_runs()
        )  # how to include: page_token=args.page, page_size=args.page_size ?
        json.dump(response, sys.stdout, indent=4)
        return 0

    if args.log:
        response = client.get_run_log(run_id=args.log)
        sys.stdout.write(requests.get(response["run_log"]["stderr"], headers=auth).text)
        return 0

    if args.get:
        response = client.get_run_log(run_id=args.get)
        json.dump(response, sys.stdout, indent=4)
        return 0

    if args.info:
        response = client.get_service_info()
        json.dump(response, sys.stdout, indent=4)
        return 0

    if not args.workflow_url:
        parser.print_help()
        return 1

    if not args.job_order:
        logging.error("Missing json/yaml file.")
        return 1

    job_order = modify_jsonyaml_paths(args.job_order)

    if args.quiet:
        logging.basicConfig(level=logging.WARNING)
    else:
        logging.basicConfig(level=logging.INFO)

    args.attachments = "" if not args.attachments else args.attachments.split(",")
    r = client.run(args.workflow_url, job_order, args.attachments)

    if args.wait:
        logging.info("Workflow run id is %s", r["run_id"])
    else:
        sys.stdout.write(r["run_id"] + "\n")
        exit(0)

    r = client.get_run_status(run_id=r["run_id"])
    while r["state"] in ("QUEUED", "INITIALIZING", "RUNNING"):
        time.sleep(8)
        r = client.get_run_status(run_id=r["run_id"])

    logging.info("State is %s", r["state"])

    s = client.get_run_log(run_id=r["run_id"])

    try:
        # TODO: Only works with Arvados atm
        logging.info(str(s["run_log"]["stderr"]))
        logs = requests.get(s["run_log"]["stderr"], headers=auth).text
        logging.info("Run log:\n" + logs)
    except InvalidSchema:
        logging.info("Run log:\n" + str(s["run_log"]["stderr"]))
    except MissingSchema:
        logging.info("Run log:\n" + str(s["run_log"]["stderr"]))

    # print the output json
    if "fields" in s["outputs"] and s["outputs"]["fields"] is None:
        del s["outputs"]["fields"]
    json.dump(s["outputs"], sys.stdout, indent=4)

    if r["state"] == "COMPLETE":
        return 0
    else:
        return 1


if __name__ == "__main__":
    sys.exit(main(sys.argv[1:]))



import tempfile
import json
import os
import logging

import connexion
from werkzeug.utils import secure_filename


def visit(d, op):
    """Recursively call op(d) for all list subelements and dictionary 'values' that d may have."""
    op(d)
    if isinstance(d, list):
        for i in d:
            visit(i, op)
    elif isinstance(d, dict):
        for i in d.values():
            visit(i, op)


class WESBackend:
    """Stores and retrieves options.  Intended to be inherited."""

    def __init__(self, opts):
        """Parse and store options as a list of tuples."""
        self.pairs = []
        for o in opts if opts else []:
            k, v = o.split("=", 1)
            self.pairs.append((k, v))

    def getopt(self, p, default=None):
        """Returns the first option value stored that matches p or default."""
        for k, v in self.pairs:
            if k == p:
                return v
        return default

    def getoptlist(self, p):
        """Returns all option values stored that match p as a list."""
        optlist = []
        for k, v in self.pairs:
            if k == p:
                optlist.append(v)
        return optlist

    def log_for_run(self, run_id, message):
        logging.info("Workflow %s: %s", run_id, message)

    def collect_attachments(self, run_id=None):
        tempdir = tempfile.mkdtemp()
        body = {}
        has_attachments = False
        for k, ls in connexion.request.files.lists():
            try:
                for v in ls:
                    if k == "workflow_attachment":
                        sp = v.filename.split("/")
                        fn = []
                        for p in sp:
                            if p not in ("", ".", ".."):
                                fn.append(secure_filename(p))
                        dest = os.path.join(tempdir, *fn)
                        if not os.path.isdir(os.path.dirname(dest)):
                            os.makedirs(os.path.dirname(dest))
                        self.log_for_run(
                            run_id,
                            f"Staging attachment '{v.filename}' to '{dest}'",
                        )
                        v.save(dest)
                        has_attachments = True
                        body[k] = (
                            "file://%s" % tempdir
                        )  # Reference to temp working dir.
                    elif k in ("workflow_params", "tags", "workflow_engine_parameters"):
                        content = v.read()
                        body[k] = json.loads(content.decode("utf-8"))
                    else:
                        body[k] = v.read().decode()
            except Exception as e:
                raise ValueError(f"Error reading parameter '{k}': {e}")
        for k, ls in connexion.request.form.lists():
            try:
                for v in ls:
                    if not v:
                        continue
                    if k in ("workflow_params", "tags", "workflow_engine_parameters"):
                        body[k] = json.loads(v)
                    else:
                        body[k] = v
            except Exception as e:
                raise ValueError(f"Error reading parameter '{k}': {e}")

        if "workflow_url" in body:
            if ":" not in body["workflow_url"]:
                if not has_attachments:
                    raise ValueError(
                        "Relative 'workflow_url' but missing 'workflow_attachment'"
                    )
                body["workflow_url"] = "file://%s" % os.path.join(
                    tempdir, secure_filename(body["workflow_url"])
                )
            self.log_for_run(
                run_id, "Using workflow_url '%s'" % body.get("workflow_url")
            )
        else:
            raise ValueError("Missing 'workflow_url' in submission")

        if "workflow_params" not in body:
            raise ValueError("Missing 'workflow_params' in submission")

        return tempdir, body



import unittest
import os
import logging
import subprocess
import sys

pkg_root = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))  # noqa
sys.path.insert(0, pkg_root)  # noqa

from wes_client.util import expand_globs, wf_info

logging.basicConfig(level=logging.INFO)


class IntegrationTest(unittest.TestCase):
    def setUp(self):
        dirname, filename = os.path.split(os.path.abspath(__file__))
        self.testdata_dir = dirname + "data"
        self.local = {
            "cwl": "file://" + os.path.join(os.getcwd() + "/testdata/md5sum.cwl"),
            "wdl": "file://" + os.path.join(os.getcwd() + "/testdata/md5sum.wdl"),
            "py": "file://" + os.path.join(os.getcwd() + "/test/test_integration.py"),
            "unsupported": "fake.txt",
        }

        self.remote = {
            "cwl": "https://raw.githubusercontent.com/common-workflow-language/workflow-service/master/testdata/md5sum.cwl",
            "wdl": "https://raw.githubusercontent.com/common-workflow-language/workflow-service/master/testdata/md5sum.wdl",
            "py": "https://raw.githubusercontent.com/common-workflow-language/workflow-service/master/test/test_integration.py",
            "unsupported": "gs://topmed_workflow_testing/topmed_aligner/small_test_files_sbg/example_human_known_snp.py",
            "unreachable": "https://fake.py",
        }

        self.expected = {
            "cwl": ("v1.0", "CWL"),
            "wdl": ("draft-2", "WDL"),
            "py": ("2.7", "PY"),
            "pyWithPrefix": ("2.7", "PY"),
        }

    def tearDown(self):
        unittest.TestCase.tearDown(self)

    def test_expand_globs(self):
        """Asserts that wes_client.expand_globs() sees the same files in the cwd as 'ls'."""
        files = subprocess.check_output(["ls", "-1", "."])

        # python 2/3 bytestring/utf-8 compatibility
        if isinstance(files, str):
            files = files.split("\n")
        else:
            files = files.decode("utf-8").split("\n")

        if "" in files:
            files.remove("")
        files = ["file://" + os.path.abspath(f) for f in files]
        glob_files = expand_globs("*")
        assert set(files) == glob_files, "\n" + str(set(files)) + "\n" + str(glob_files)

    def testSupportedFormatChecking(self):
        """
        Check that non-wdl, -python, -cwl files are rejected.

        This test is run only on local files to avoid downloading and removing a new file.
        """

        for file_format, location in self.local.items():
            if file_format != "unsupported":
                # Tests the behavior after receiving supported file types with and without the 'file://' prefix
                self.assertEqual(wf_info(location), self.expected[file_format])
                self.assertEqual(wf_info(location[7:]), self.expected[file_format])

            else:
                # Tests behavior after receiving a non supported file type.
                with self.assertRaises(TypeError):
                    wf_info(location)

    def testFileLocationChecking(self):
        """
        Check that the function rejects unsupported file locations.

        This test needs to be run on remote files to test the location checking functionality of wf_info().
        """

        for file_format, location in self.remote.items():
            if file_format == "unsupported":
                # Tests behavior after receiving a file hosted at an unsupported location.
                with self.assertRaises(NotImplementedError):
                    wf_info(location)

            elif file_format == "unreachable":
                # Tests behavior after receiving a non-existent file.
                with self.assertRaises(IOError):
                    wf_info(location)

            else:
                self.assertEqual(wf_info(location), self.expected[file_format])
                self.assertFalse(
                    os.path.isfile(
                        os.path.join(os.getcwd(), "fetchedFromRemote." + file_format)
                    )
                )


if __name__ == "__main__":
    unittest.main()  # run all tests



#!/usr/bin/env python

import os
from setuptools import setup

SETUP_DIR = os.path.dirname(__file__)

long_description = ""

with open("README.pypi.rst") as readmeFile:
    long_description = readmeFile.read()

setup(
    name="wes-service",
    version="4.0",
    description="GA4GH Workflow Execution Service reference implementation",
    long_description=long_description,
    author="GA4GH Containers and Workflows task team",
    author_email="common-workflow-language@googlegroups.com",
    url="https://github.com/common-workflow-language/cwltool-service",
    download_url="https://github.com/common-workflow-language/cwltool-service",
    license="Apache 2.0",
    python_requires="~=3.7",
    setup_requires=["pytest-runner"],
    tests_require=["pytest"],
    packages=["wes_service", "wes_client"],
    package_data={"wes_service": ["openapi/workflow_execution_service.swagger.yaml"]},
    include_package_data=True,
    install_requires=[
        "connexion >= 2.0.2, < 3",
        "ruamel.yaml >= 0.15.78",
        "schema-salad",
    ],
    entry_points={
        "console_scripts": [
            "wes-server=wes_service.wes_service_main:main",
            "wes-client=wes_client.wes_client_main:main",
        ]
    },
    extras_require={
        "cwltool": ["cwlref-runner"],
        "arvados": ["arvados-cwl-runner"],
        "toil": ["toil[cwl]==5.6.0"],
    },
    zip_safe=False,
    platforms=["MacOS X", "Posix"],
    classifiers=[
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Topic :: Software Development :: Libraries :: Python Modules",
    ],
)



import unittest
import time
import os
import subprocess
import signal
import shutil
import logging
import sys
import requests
import pytest

pkg_root = os.path.abspath(os.path.join(os.path.dirname(__file__), ".."))  # noqa
sys.path.insert(0, pkg_root)  # noqa

from wes_client.util import WESClient

logging.basicConfig(level=logging.INFO)


class IntegrationTest(unittest.TestCase):
    """A baseclass that's inherited for use with different cwl backends."""

    @classmethod
    def setUpClass(cls):
        # cwl
        cls.cwl_dockstore_url = "https://dockstore.org/api/ga4gh/trs/v2/tools/quay.io%2Fbriandoconnor%2Fdockstore-tool-md5sum/versions/1.0.4/plain-CWL/descriptor//Dockstore.cwl"
        cls.cwl_local_path = "file://" + os.path.abspath("testdata/md5sum.cwl")
        cls.cwl_json_input = "file://" + os.path.abspath("testdata/md5sum.json")
        cls.cwl_attachments = [
            "file://" + os.path.abspath("testdata/md5sum.input"),
            "file://" + os.path.abspath("testdata/dockstore-tool-md5sum.cwl"),
        ]
        # wdl
        cls.wdl_local_path = os.path.abspath("testdata/md5sum.wdl")
        cls.wdl_json_input = "file://" + os.path.abspath("testdata/md5sum.wdl.json")
        cls.wdl_attachments = ["file://" + os.path.abspath("testdata/md5sum.input")]

        # client for the swagger API methods
        cls.client = WESClient(
            {"auth": {"Authorization": ""}, "proto": "http", "host": "localhost:8080"}
        )

        # manual test (wdl only working locally atm)
        cls.manual = False

    def setUp(self):
        """Start a (local) wes-service server to make requests against."""
        raise NotImplementedError

    def tearDown(self):
        """Kill the wes-service server."""
        os.kill(self.wes_server_process.pid, signal.SIGTERM)
        while get_server_pids():
            for pid in get_server_pids():
                try:
                    os.kill(int(pid), signal.SIGKILL)
                    time.sleep(3)
                except OSError as e:
                    print(e)
        unittest.TestCase.tearDown(self)

    def test_dockstore_md5sum(self):
        """HTTP md5sum cwl (dockstore), run it on the wes-service server, and check for the correct output."""
        outfile_path, run_id = self.run_md5sum(
            wf_input=self.cwl_dockstore_url,
            json_input=self.cwl_json_input,
            workflow_attachment=self.cwl_attachments,
        )
        state = self.wait_for_finish(run_id)
        self.check_complete(run_id)
        self.assertTrue(
            self.check_for_file(outfile_path),
            "Output file was not found: " + str(outfile_path),
        )

    def test_local_md5sum(self):
        """LOCAL md5sum cwl to the wes-service server, and check for the correct output."""
        outfile_path, run_id = self.run_md5sum(
            wf_input=self.cwl_local_path,
            json_input=self.cwl_json_input,
            workflow_attachment=self.cwl_attachments,
        )
        state = self.wait_for_finish(run_id)
        self.check_complete(run_id)
        self.assertTrue(
            self.check_for_file(outfile_path),
            "Output file was not found: " + str(outfile_path),
        )

    # See https://ga4gh.github.io/workflow-execution-service-schemas/docs/#operation/GetRunLog
    @pytest.mark.skip(
        "workflow_attachment is not part of WES spec for the log.request body"
    )
    def test_run_attachments(self):
        """LOCAL md5sum cwl to the wes-service server, check for attachments."""
        outfile_path, run_id = self.run_md5sum(
            wf_input=self.cwl_local_path,
            json_input=self.cwl_json_input,
            workflow_attachment=self.cwl_attachments,
        )
        get_response = self.client.get_run_log(run_id)["request"]
        state = self.wait_for_finish(run_id)
        self.check_complete(run_id)
        self.assertTrue(
            self.check_for_file(outfile_path),
            "Output file was not found: " + get_response["workflow_attachment"],
        )
        attachment_tool_path = (
            get_response["workflow_attachment"][7:] + "/dockstore-tool-md5sum.cwl"
        )
        self.assertTrue(
            self.check_for_file(attachment_tool_path),
            "Attachment file was not found: " + get_response["workflow_attachment"],
        )

    def test_get_service_info(self):
        """
        Test wes_client.util.WESClient.get_service_info()

        This method will exit(1) if the response is not 200.
        """
        r = self.client.get_service_info()
        assert "workflow_type_versions" in r
        assert "supported_wes_versions" in r
        assert "supported_filesystem_protocols" in r
        assert "workflow_engine_versions" in r

    def test_list_runs(self):
        """
        Test wes_client.util.WESClient.list_runs()

        This method will exit(1) if the response is not 200.
        """
        r = self.client.list_runs()
        assert "workflows" in r

    def test_get_run_status(self):
        """
        Test wes_client.util.WESClient.run_status()

        This method will exit(1) if the response is not 200.
        """
        outfile_path, run_id = self.run_md5sum(
            wf_input=self.cwl_local_path,
            json_input=self.cwl_json_input,
            workflow_attachment=self.cwl_attachments,
        )
        r = self.client.get_run_status(run_id)
        assert "state" in r
        assert "run_id" in r

    def run_md5sum(self, wf_input, json_input, workflow_attachment=None):
        """Pass a local md5sum cwl to the wes-service server, and return the path of the output file that was created."""
        response = self.client.run(wf_input, json_input, workflow_attachment)
        assert "run_id" in response, str(response.json())
        output_dir = os.path.abspath(
            os.path.join("workflows", response["run_id"], "outdir")
        )
        return os.path.join(output_dir, "md5sum.txt"), response["run_id"]

    def wait_for_finish(self, run_id, seconds=120):
        """Return True if a file exists within a certain amount of time."""
        wait_counter = 0
        r = self.client.get_run_status(run_id)
        while r["state"] in ("QUEUED", "INITIALIZING", "RUNNING"):
            time.sleep(1)
            wait_counter += 1
            if wait_counter > seconds:
                return None
            r = self.client.get_run_status(run_id)
        return r["state"]

    def check_complete(self, run_id):
        s = self.client.get_run_log(run_id)
        if s["state"] != "COMPLETE":
            logging.info(str(s["run_log"]["stderr"]))
            if str(s["run_log"]["stderr"]).startswith("http"):
                logs = requests.get(
                    s["run_log"]["stderr"], headers=self.client.auth
                ).text
                logging.info("Run log:\n" + logs)
        assert s["state"] == "COMPLETE"

    def check_for_file(self, filepath, seconds=120):
        """Return True if a file exists within a certain amount of time."""
        wait_counter = 0
        while not os.path.exists(filepath):
            time.sleep(1)
            wait_counter += 1
            if wait_counter > seconds:
                return False
        return True


def get_server_pids():
    try:
        pids = (
            subprocess.check_output(["pgrep", "-f", "wes_service_main.py"])
            .strip()
            .split()
        )
    except subprocess.CalledProcessError:
        return None
    return pids


class CwltoolTest(IntegrationTest):
    """Test using cwltool."""

    def setUp(self):
        """
        Start a (local) wes-service server to make requests against.
        Use cwltool as the wes-service server 'backend'.
        """
        if os.path.exists("workflows"):
            shutil.rmtree("workflows")
        self.wes_server_process = subprocess.Popen(
            [
                "python",
                os.path.abspath("wes_service/wes_service_main.py"),
                "--backend=wes_service.cwl_runner",
                "--opt",
                "runner=cwltool",
                "--port=8080",
                "--debug",
            ]
        )
        time.sleep(5)


class ToilTest(IntegrationTest):
    """Test using Toil."""

    def setUp(self):
        """
        Start a (local) wes-service server to make requests against.
        Use toil as the wes-service server 'backend'.
        """
        self.wes_server_process = subprocess.Popen(
            "python {} --backend=wes_service.toil_wes "
            '--opt="extra=--logLevel=CRITICAL" '
            '--opt="extra=--clean=never"'
            "".format(os.path.abspath("wes_service/wes_service_main.py")),
            shell=True,
        )
        time.sleep(5)

    def test_local_wdl(self):
        """LOCAL md5sum wdl to the wes-service server, and check for the correct output."""
        # Working locally but not on travis... >.<;
        if self.manual:
            outfile_path, run_id = self.run_md5sum(
                wf_input=self.wdl_local_path,
                json_input=self.wdl_json_input,
                workflow_attachment=self.wdl_attachments,
            )
            self.assertTrue(
                self.check_for_file(outfile_path),
                "Output file was not found: " + str(outfile_path),
            )


@pytest.mark.skipif(
    not os.environ.get("ARVADOS_API_TOKEN"), reason="Arvados not configured"
)
class ArvadosTest(IntegrationTest):
    """Test using arvados-cwl-runner."""

    def setUp(self):
        """
        Start a (local) wes-service server to make requests against.
        Use arvados-cwl-runner as the wes-service server 'backend'.
        Requires ARVADOS_API_HOST and ARVADOS_API_TOKEN to be set in the environment.
        """
        if os.path.exists("workflows"):
            shutil.rmtree("workflows")
        self.wes_server_process = subprocess.Popen(
            [
                "python",
                os.path.abspath("wes_service/wes_service_main.py"),
                "--backend=wes_service.arvados_wes",
                "--port=8080",
                "--debug",
            ]
        )
        self.client.auth = {
            "Authorization": "Bearer " + os.environ["ARVADOS_API_TOKEN"]
        }
        time.sleep(5)

    def check_for_file(self, filepath, seconds=120):
        # Doesn't make sense for arvados
        return True


# Prevent pytest/unittest's discovery from attempting to discover the base test class.
del IntegrationTest


if __name__ == "__main__":
    unittest.main()  # run all tests



import os
import json
import schema_salad.ref_resolver
from subprocess import check_call, DEVNULL, CalledProcessError
import yaml
import glob
import requests
import logging

from wes_service.util import visit

from urllib.request import urlopen, pathname2url


def two_seven_compatible(filePath):
    """Determines if a python file is 2.7 compatible by seeing if it compiles in a subprocess"""
    try:
        check_call(["python2", "-m", "py_compile", filePath], stderr=DEVNULL)
    except CalledProcessError:
        raise RuntimeError("Python files must be 2.7 compatible")
    return True


def get_version(extension, workflow_file):
    """Determines the version of a .py, .wdl, or .cwl file."""
    if extension == "py" and two_seven_compatible(workflow_file):
        return "2.7"
    elif extension == "cwl":
        return yaml.load(open(workflow_file), Loader=yaml.FullLoader)["cwlVersion"]
    else:  # Must be a wdl file.
        # Borrowed from https://github.com/Sage-Bionetworks/synapse-orchestrator/blob/develop/synorchestrator/util.py#L142
        try:
            return [
                entry.lstrip("version")
                for entry in workflow_file.splitlines()
                if "version" in entry.split(" ")
            ][0]
        except IndexError:
            return "draft-2"


def wf_info(workflow_path):
    """
    Returns the version of the file and the file extension.

    Assumes that the file path is to the file directly ie, ends with a valid file extension.Supports checking local
    files as well as files at http:// and https:// locations. Files at these remote locations are recreated locally to
    enable our approach to version checking, then removed after version is extracted.
    """

    supported_formats = ["py", "wdl", "cwl"]
    file_type = workflow_path.lower().split(".")[-1]  # Grab the file extension
    workflow_path = workflow_path if ":" in workflow_path else "file://" + workflow_path

    if file_type in supported_formats:
        if workflow_path.startswith("file://"):
            version = get_version(file_type, workflow_path[7:])
        elif workflow_path.startswith("https://") or workflow_path.startswith(
            "http://"
        ):
            # If file not local go fetch it.
            html = urlopen(workflow_path).read()
            local_loc = os.path.join(os.getcwd(), "fetchedFromRemote." + file_type)
            with open(local_loc, "w") as f:
                f.write(html.decode())
            version = wf_info("file://" + local_loc)[
                0
            ]  # Don't take the file_type here, found it above.
            os.remove(
                local_loc
            )  # TODO: Find a way to avoid recreating file before version determination.
        else:
            raise NotImplementedError(
                "Unsupported workflow file location: {}. Must be local or HTTP(S).".format(
                    workflow_path
                )
            )
    else:
        raise TypeError(
            "Unsupported workflow type: .{}. Must be {}.".format(
                file_type, ".py, .cwl, or .wdl"
            )
        )
    return version, file_type.upper()


def modify_jsonyaml_paths(jsonyaml_file):
    """
    Changes relative paths in a json/yaml file to be relative
    to where the json/yaml file is located.

    :param jsonyaml_file: Path to a json/yaml file.
    """
    loader = schema_salad.ref_resolver.Loader(
        {"location": {"@type": "@id"}, "path": {"@type": "@id"}}
    )
    input_dict, _ = loader.resolve_ref(jsonyaml_file, checklinks=False)
    basedir = os.path.dirname(jsonyaml_file)

    def fixpaths(d):
        """Make sure all paths have a URI scheme."""
        if isinstance(d, dict):
            if "path" in d:
                if ":" not in d["path"]:
                    local_path = os.path.normpath(
                        os.path.join(os.getcwd(), basedir, d["path"])
                    )
                    d["location"] = pathname2url(local_path)
                else:
                    d["location"] = d["path"]
                del d["path"]

    visit(input_dict, fixpaths)
    return json.dumps(input_dict)


def build_wes_request(workflow_file, json_path, attachments=None):
    """
    :param str workflow_file: Path to cwl/wdl file.  Can be http/https/file.
    :param json_path: Path to accompanying json file.
    :param attachments: Any other files needing to be uploaded to the server.

    :return: A list of tuples formatted to be sent in a post to the wes-server (Swagger API).
    """
    workflow_file = (
        "file://" + workflow_file if ":" not in workflow_file else workflow_file
    )
    wfbase = None
    if json_path.startswith("file://"):
        wfbase = os.path.dirname(json_path[7:])
        json_path = json_path[7:]
        with open(json_path) as f:
            wf_params = json.dumps(json.load(f))
    elif json_path.startswith("http"):
        wf_params = modify_jsonyaml_paths(json_path)
    else:
        wf_params = json_path
    wf_version, wf_type = wf_info(workflow_file)

    parts = [
        ("workflow_params", wf_params),
        ("workflow_type", wf_type),
        ("workflow_type_version", wf_version),
    ]

    if workflow_file.startswith("file://"):
        if wfbase is None:
            wfbase = os.path.dirname(workflow_file[7:])
        parts.append(
            (
                "workflow_attachment",
                (os.path.basename(workflow_file[7:]), open(workflow_file[7:], "rb")),
            )
        )
        parts.append(("workflow_url", os.path.basename(workflow_file[7:])))
    else:
        parts.append(("workflow_url", workflow_file))

    if wfbase is None:
        wfbase = os.getcwd()
    if attachments:
        for attachment in attachments:
            if attachment.startswith("file://"):
                attachment = attachment[7:]
                attach_f = open(attachment, "rb")
                relpath = os.path.relpath(attachment, wfbase)
            elif attachment.startswith("http"):
                attach_f = urlopen(attachment)
                relpath = os.path.basename(attach_f)

            parts.append(("workflow_attachment", (relpath, attach_f)))

    return parts


def expand_globs(attachments):
    expanded_list = []
    for filepath in attachments:
        if "file://" in filepath:
            for f in glob.glob(filepath[7:]):
                expanded_list += ["file://" + os.path.abspath(f)]
        elif ":" not in filepath:
            for f in glob.glob(filepath):
                expanded_list += ["file://" + os.path.abspath(f)]
        else:
            expanded_list += [filepath]
    return set(expanded_list)


def wes_reponse(postresult):
    if postresult.status_code != 200:
        error = str(json.loads(postresult.text))
        logging.error(error)
        raise Exception(error)

    return json.loads(postresult.text)


class WESClient:
    def __init__(self, service):
        self.auth = service["auth"]
        self.proto = service["proto"]
        self.host = service["host"]

    def get_service_info(self):
        """
        Get information about Workflow Execution Service. May
        include information related (but not limited to) the
        workflow descriptor formats, versions supported, the
        WES API versions supported, and information about general
        the service availability.

        :param str auth: String to send in the auth header.
        :param proto: Schema where the server resides (http, https)
        :param host: Port where the post request will be sent and the wes server listens at (default 8080)
        :return: The body of the get result as a dictionary.
        """
        postresult = requests.get(
            f"{self.proto}://{self.host}/ga4gh/wes/v1/service-info",
            headers=self.auth,
        )
        return wes_reponse(postresult)

    def list_runs(self):
        """
        List the workflows, this endpoint will list the workflows
        in order of oldest to newest. There is no guarantee of
        live updates as the user traverses the pages, the behavior
        should be decided (and documented) by each implementation.

        :param str auth: String to send in the auth header.
        :param proto: Schema where the server resides (http, https)
        :param host: Port where the post request will be sent and the wes server listens at (default 8080)
        :return: The body of the get result as a dictionary.
        """
        postresult = requests.get(
            f"{self.proto}://{self.host}/ga4gh/wes/v1/runs", headers=self.auth
        )
        return wes_reponse(postresult)

    def run(self, wf, jsonyaml, attachments):
        """
        Composes and sends a post request that signals the wes server to run a workflow.

        :param str workflow_file: A local/http/https path to a cwl/wdl/python workflow file.
        :param str jsonyaml: A local path to a json or yaml file.
        :param list attachments: A list of local paths to files that will be uploaded to the server.
        :param str auth: String to send in the auth header.
        :param proto: Schema where the server resides (http, https)
        :param host: Port where the post request will be sent and the wes server listens at (default 8080)

        :return: The body of the post result as a dictionary.
        """
        attachments = list(expand_globs(attachments))
        parts = build_wes_request(wf, jsonyaml, attachments)
        postresult = requests.post(
            f"{self.proto}://{self.host}/ga4gh/wes/v1/runs",
            files=parts,
            headers=self.auth,
        )
        return wes_reponse(postresult)

    def cancel(self, run_id):
        """
        Cancel a running workflow.

        :param run_id: String (typically a uuid) identifying the run.
        :param str auth: String to send in the auth header.
        :param proto: Schema where the server resides (http, https)
        :param host: Port where the post request will be sent and the wes server listens at (default 8080)
        :return: The body of the delete result as a dictionary.
        """
        postresult = requests.post(
            f"{self.proto}://{self.host}/ga4gh/wes/v1/runs/{run_id}/cancel",
            headers=self.auth,
        )
        return wes_reponse(postresult)

    def get_run_log(self, run_id):
        """
        Get detailed info about a running workflow.

        :param run_id: String (typically a uuid) identifying the run.
        :param str auth: String to send in the auth header.
        :param proto: Schema where the server resides (http, https)
        :param host: Port where the post request will be sent and the wes server listens at (default 8080)
        :return: The body of the get result as a dictionary.
        """
        postresult = requests.get(
            f"{self.proto}://{self.host}/ga4gh/wes/v1/runs/{run_id}",
            headers=self.auth,
        )
        return wes_reponse(postresult)

    def get_run_status(self, run_id):
        """
        Get quick status info about a running workflow.

        :param run_id: String (typically a uuid) identifying the run.
        :param str auth: String to send in the auth header.
        :param proto: Schema where the server resides (http, https)
        :param host: Port where the post request will be sent and the wes server listens at (default 8080)
        :return: The body of the get result as a dictionary.
        """
        postresult = requests.get(
            f"{self.proto}://{self.host}/ga4gh/wes/v1/runs/{run_id}/status",
            headers=self.auth,
        )
        return wes_reponse(postresult)
